#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API - Applier."""


import logging

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.utils import IntegrityError
from rest_framework.exceptions import ValidationError as ValidationError

from .common import NON_FIELD_ERRORS, Change, ChangeSet, ChangeSetException, ChangeSetResult, ChangeType, error_from_validation_error
from .matcher import find_existing_object
from .plugin_utils import get_object_type_model, legal_fields
from .supported_models import get_serializer_for_model

logger = logging.getLogger(__name__)


def apply_changeset(change_set: ChangeSet, request) -> ChangeSetResult:
    """Apply a change set."""
    _validate_change_set(change_set)

    created = {}
    for change in change_set.changes:
        change_type = change.change_type
        object_type = change.object_type

        if change_type == ChangeType.NOOP.value:
            continue

        try:
            model_class = get_object_type_model(object_type)
            data = _pre_apply(model_class, change, created)
            _apply_change(data, model_class, change, created, request)
        except ValidationError as e:
            raise error_from_validation_error(e, object_type)
        except ObjectDoesNotExist:
            raise _err(f"{object_type} with id {change.object_id} does not exist", object_type, "object_id")
        except TypeError as e:
            # this indicates a problem in model validation (should raise ValidationError)
            # but raised non-validation error (TypeError) -- we don't know which field trigged it.
            import traceback
            traceback.print_exc()
            logger.error(f"validation raised TypeError error on unspecified field of {object_type}: {data}: {e}")
            logger.error(traceback.format_exc())
            raise _err("invalid data type for field (TypeError)", object_type, "__all__")
        except IntegrityError as e:
            logger.error(f"Integrity error {object_type}: {e} {data}")
            raise _err(f"created a conflict with an existing {object_type}", object_type, "__all__")

    return ChangeSetResult(
        id=change_set.id,
    )

def _is_auto_created_component(object_type: str) -> bool:
    """Check if the object type is auto-created from templates."""
    auto_created_components = [
        "dcim.consoleport",
        "dcim.consoleserverport",
        "dcim.powerport",
        "dcim.poweroutlet",
        "dcim.interface",
        "dcim.rearport",
        "dcim.frontport",
        "dcim.modulebay",
        "dcim.devicebay",
        "dcim.inventoryitem",
    ]
    return object_type in auto_created_components


def _try_find_and_update_existing_instance(data: dict, object_type: str, serializer_class, request):
    """Try to find existing auto-created instance and update it."""
    try:
        instance = find_existing_object(data, object_type)
        if instance:
            serializer = serializer_class(instance, data=data, partial=True, context={"request": request})
            serializer.is_valid(raise_exception=True)
            return serializer.save()
    except (ValueError, TypeError) as e:
        logger.debug(f"Could not find existing {object_type}: {e}")
    return None


def _create_or_find_instance(data: dict, object_type: str, serializer_class, request):
    """Create new instance or find existing one on conflict."""
    serializer = serializer_class(data=data, context={"request": request})
    try:
        serializer.is_valid(raise_exception=True)
        return serializer.save()
    except ValidationError as e:
        instance = find_existing_object(data, object_type)
        if not instance:
            raise e
        return instance


def _apply_change(data: dict, model_class: models.Model, change: Change, created: dict, request):
    serializer_class = get_serializer_for_model(model_class)
    change_type = change.change_type

    if change_type == ChangeType.CREATE.value:
        # For component types that may be auto-created from e.g. DeviceType or ModuleType templates,
        # try to find existing object first before attempting to create.
        # This prevents duplicates when components are instantiated during Device/Module save()
        instance = None
        if _is_auto_created_component(change.object_type):
            instance = _try_find_and_update_existing_instance(data, change.object_type, serializer_class, request)

        if not instance:
            instance = _create_or_find_instance(data, change.object_type, serializer_class, request)

        # Always add the instance to created dict so it can be referenced by subsequent changes
        if change.ref_id:
            created[change.ref_id] = instance

    elif change_type == ChangeType.UPDATE.value:
        if object_id := change.object_id:
            instance = model_class.objects.get(id=object_id)
            serializer = serializer_class(instance, data=data, partial=True, context={"request": request})
            serializer.is_valid(raise_exception=True)
            serializer.save()
        # create and update in a same change set
        elif change.ref_id and (instance := created[change.ref_id]):
            serializer = serializer_class(instance, data=data, partial=True, context={"request": request})
            serializer.is_valid(raise_exception=True)
            serializer.save()

def _set_path(data, path, value):
    path = path.split(".")
    key = path.pop(0)
    while len(path) > 0:
        data = data[key]
        key = path.pop(0)
    data[key] = value

def _get_path(data, path):
    path = path.split(".")
    v = data
    for p in path:
        v = v[p]
    return v

def _pre_apply(model_class: models.Model, change: Change, created: dict):
    data = change.data.copy()

    # resolve foreign key references to new objects
    for ref_field in change.new_refs:
        v = _get_path(data, ref_field)
        if isinstance(v, list | tuple):
            ref_list = []
            for ref in v:
                if isinstance(ref, str):
                    ref_list.append(created[ref].pk)
                elif isinstance(ref, int):
                    ref_list.append(ref)
            _set_path(data, ref_field, ref_list)
        else:
            _set_path(data, ref_field, created[v].pk)

    # ignore? fields that are not in the data model (error?)
    allowed_fields = legal_fields(model_class)
    for key in list(data.keys()):
        if key not in allowed_fields:
            if key != "id":
                logger.warning(f"Field {key} is not in the diode data model, ignoring.")
            data.pop(key)

    return data

def _validate_change_set(change_set: ChangeSet):
    if not change_set.id:
        raise _err("Change set ID is required", "changeset","id")
    if not change_set.changes:
        raise _err("Changes are required", "changeset", "changes")

    for change in change_set.changes:
        if change.object_id is None and change.ref_id is None:
            raise _err("Object ID or Ref ID must be provided", change.object_type, NON_FIELD_ERRORS)
        if change.change_type not in [ct.value for ct in ChangeType]:
            raise _err(f"Unsupported change type '{change.change_type}'", change.object_type, "change_type")

def _err(message, object_name, field):
    if not object_name:
        object_name = "__all__"
    return ChangeSetException(message, errors={object_name: {field: [message]}})




#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Serializers."""

from netbox.api.serializers import NetBoxModelSerializer

from netbox_diode_plugin.models import Setting


class SettingSerializer(NetBoxModelSerializer):
    """Setting Serializer."""

    class Meta:
        """Meta class."""

        model = Setting
        fields = (
            "id",
            "diode_target",
            "custom_fields",
            "created",
            "last_updated",
        )



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Database migrations."""

import utilities.json
from django.db import migrations, models
from netbox.plugins import get_plugin_config


def create_settings_entity(apps, schema_editor):
    """Create a Setting entity."""
    Setting = apps.get_model("netbox_diode_plugin", "Setting")

    default_diode_target = get_plugin_config("netbox_diode_plugin", "diode_target")
    diode_target = get_plugin_config(
        "netbox_diode_plugin", "diode_target_override", default_diode_target
    )

    Setting.objects.create(diode_target=diode_target)


class Migration(migrations.Migration):
    """Initial migration."""

    replaces = [
        ("netbox_diode_plugin", "0001_initial"),
        ("netbox_diode_plugin", "0002_setting"),
        ("netbox_diode_plugin", "0003_clear_permissions"),
        ("netbox_diode_plugin", "0004_rename_legacy_users"),
        ("netbox_diode_plugin", "0005_revoke_superuser_status"),
    ]

    initial = True

    dependencies = [
        ("contenttypes", "0001_initial"),
        ("users", "0006_custom_group_model"),
    ]

    operations = [
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True, null=True)),
                ("last_updated", models.DateTimeField(auto_now=True, null=True)),
                (
                    "custom_field_data",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        encoder=utilities.json.CustomFieldJSONEncoder,
                    ),
                ),
                ("diode_target", models.CharField(max_length=255)),
            ],
            options={
                "verbose_name": "Setting",
                "verbose_name_plural": "Diode Settings",
            },
        ),
        migrations.RunPython(
            code=create_settings_entity, reverse_code=migrations.RunPython.noop
        ),
    ]



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API Permissions."""

from rest_framework.permissions import BasePermission

SCOPE_NETBOX_READ = "netbox:read"
SCOPE_NETBOX_WRITE = "netbox:write"


class IsAuthenticated(BasePermission):
    """Check if the request is authenticated."""

    def has_permission(self, request, view):
        """Check if the request is authenticated."""
        return bool(getattr(request.user, "is_authenticated", False))


def require_scopes(*required_scopes):
    """Require one or more OAuth2 token scopes to access a view."""

    class ScopedPermission(BasePermission):
        """Check if the request has the required scopes."""

        def has_permission(self, request, view):
            """Check if the request has the required scopes."""
            scopes = getattr(request, "token_scopes", [])
            return all(scope in scopes for scope in required_scopes)

    ScopedPermission.__name__ = f"RequireScopes_{'_'.join(required_scopes)}"
    return ScopedPermission



# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Forms."""
from django import forms
from django.utils.translation import gettext_lazy as _
from netbox.plugins import get_plugin_config
from utilities.forms.rendering import FieldSet

from netbox_diode_plugin.models import Setting

__all__ = (
    "SettingsForm",
    "ClientCredentialForm",
)


class SettingsForm(forms.ModelForm):
    """Settings form."""

    # Define branch as a custom field (not part of the model directly)
    branch = forms.ModelChoiceField(
        queryset=None,
        required=False,
        label="Branch",
        help_text="Select an active branch for Diode. Leave empty to use the main schema.",
    )

    fieldsets = (
        FieldSet(
            "diode_target",
            "branch",
        ),
    )

    class Meta:
        """Meta class."""

        model = Setting
        fields = ("diode_target",)  # Only include actual model fields

    def __init__(self, *args, **kwargs):
        """Initialize the form."""
        super().__init__(*args, **kwargs)

        diode_target_override = get_plugin_config(
            "netbox_diode_plugin", "diode_target_override"
        )

        if diode_target_override:
            self.fields["diode_target"].disabled = True
            self.fields["diode_target"].help_text = (
                "This field is not allowed to be modified."
            )

        # Handle branch field based on netbox_branching plugin availability
        from django.conf import settings as django_settings

        if "netbox_branching" in django_settings.PLUGINS:
            # Branching plugin is installed, configure the branch field
            try:
                from netbox_branching.models import Branch

                self.fields["branch"].queryset = Branch.objects.filter(status="ready")

                # Set initial value from branch_id
                if self.instance and self.instance.branch_id:
                    try:
                        self.fields["branch"].initial = Branch.objects.get(id=self.instance.branch_id)
                    except Branch.DoesNotExist:
                        pass
            except ImportError:
                # Plugin is in PLUGINS but not actually available, remove the field
                self.fields.pop("branch", None)
        else:
            # Branching plugin is not installed, remove the branch field
            self.fields.pop("branch", None)

    def save(self, commit=True):
        """Save the form and update branch_id."""
        instance = super().save(commit=False)

        # Update branch_id from the branch field
        if "branch" in self.cleaned_data:
            branch = self.cleaned_data["branch"]
            instance.branch_id = branch.id if branch else None

        if commit:
            instance.save()

        return instance


class ClientCredentialForm(forms.Form):
    """Form for adding client credentials."""

    client_name = forms.CharField(
        label=_("Client Name"),
        required=True,
        help_text=_("Enter a name for the client credential that will be created for authentication to the Diode ingestion service."),
        widget=forms.TextInput(attrs={"class": "form-control"}),
    )



# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Models."""
from urllib.parse import urlparse

from django.core.exceptions import ValidationError
from django.db import models
from django.urls import reverse
from utilities.querysets import RestrictedQuerySet


def diode_target_validator(target):
    """Diode target validator."""
    try:
        parsed_target = urlparse(target)

        if parsed_target.scheme not in ["grpc", "grpcs"]:
            raise ValueError("target should start with grpc:// or grpcs://")
    except ValueError as exc:
        raise ValidationError(exc)


class Setting(models.Model):
    """
    Setting model.

    Simple model without change logging, excluded from branching.
    """

    diode_target = models.CharField(max_length=255, validators=[diode_target_validator])
    branch_id = models.BigIntegerField(
        null=True,
        blank=True,
        help_text="ID of the branch for NetBox Branching plugin integration",
    )

    objects = RestrictedQuerySet.as_manager()

    class Meta:
        """Meta class."""

        verbose_name = "Settings"
        verbose_name_plural = "Settings"

    def __str__(self):
        """Return string representation."""
        return ""

    def get_absolute_url(self):
        """Return absolute URL."""
        return reverse("plugins:netbox_diode_plugin:settings")

    @property
    def branch(self):
        """
        Return the Branch object if branch_id is set and branching plugin is installed.

        Returns None if:
        - branch_id is not set
        - branching plugin is not installed
        - branch with given ID does not exist
        """
        if not self.branch_id:
            return None

        try:
            from netbox_branching.models import Branch
            return Branch.objects.get(id=self.branch_id)
        except (ImportError, Exception):
            return None

    @branch.setter
    def branch(self, branch_obj):
        """Set branch_id from a Branch object."""
        if branch_obj is None:
            self.branch_id = None
        else:
            self.branch_id = branch_obj.id

    @property
    def branch_schema_id(self):
        """Return the branch schema_id if branch is set."""
        branch = self.branch
        return branch.schema_id if branch else None


class UnmanagedModelManager(models.Manager):
    """Manager for unmanaged models that prevents database queries."""

    def get_queryset(self):
        """Return an empty queryset without hitting the database."""
        return super().get_queryset().none()


class ClientCredentials(models.Model):
    """Dummy model to allow for permissions, saved filters, etc.."""

    objects = UnmanagedModelManager()

    class Meta:
        """Meta class."""

        managed = False

        default_permissions = ()

        permissions = (
            ("view_clientcredentials", "Can view Client Credentials"),
            ("add_clientcredentials", "Can perform actions on Client Credentials"),
        )




#!/usr/bin/env python
# Copyright 2025 NetBox Labs Inc
"""Diode NetBox Plugin - API - Differ."""

import copy
import datetime
import logging
from collections import defaultdict

from django.contrib.contenttypes.models import ContentType
from rest_framework import serializers
from utilities.data import shallow_compare_dict

from .common import (
    NON_FIELD_ERRORS,
    Change,
    ChangeSet,
    ChangeSetException,
    ChangeSetResult,
    ChangeType,
    error_from_validation_error,
    harmonize_formats,
    sort_ints_first,
)
from .plugin_utils import get_primary_value, legal_fields
from .supported_models import extract_supported_models
from .transformer import cleanup_unresolved_references, set_custom_field_defaults, transform_proto_json

logger = logging.getLogger(__name__)

def prechange_data_from_instance(instance) -> dict: # noqa: C901
    """Convert model instance data to a dictionary format for comparison."""
    prechange_data = {}

    if instance is None:
        return prechange_data

    model_class = instance.__class__
    object_type = f"{model_class._meta.app_label}.{model_class._meta.model_name}"

    supported_models = extract_supported_models()
    model = supported_models.get(object_type)
    if not model:
        raise serializers.ValidationError({
            NON_FIELD_ERRORS: [f"{object_type} is not supported in this version."]
        })

    fields = model.get("fields", {})
    if not fields:
        raise serializers.ValidationError({
            NON_FIELD_ERRORS: [f"Model {model_class.__name__} has no fields"]
        })

    diode_fields = legal_fields(model_class)

    for field_name, field_info in fields.items():
        # permit only diode fields and the primary key
        if field_name not in diode_fields and field_name != "id":
            continue

        if not hasattr(instance, field_name):
            continue

        value = getattr(instance, field_name)
        if hasattr(value, "all"):  # Handle many-to-many and many-to-one relationships
            # For any relationship that has an 'all' method, get all related objects' primary keys
            prechange_data[field_name] = (
                sorted([_pk_or_content_type_ref(item) for item in value.all()] if value is not None else [])
            )
        elif hasattr(value, "pk"):
            # Handle regular related fields (ForeignKey, OneToOne) andContentType fields
            prechange_data[field_name] = _pk_or_content_type_ref(value)
        else:
            prechange_data[field_name] = value

    if hasattr(instance, "get_custom_fields"):
        custom_field_values = instance.get_custom_fields()
        cfmap = {}
        for cf, value in custom_field_values.items():
            if isinstance(value, datetime.datetime | datetime.date):
                cfmap[cf.name] = value
            else:
                cfmap[cf.name] = cf.serialize(value)
        prechange_data["custom_fields"] = cfmap
    prechange_data = harmonize_formats(prechange_data)

    return prechange_data

def _pk_or_content_type_ref(value):
    if isinstance(value, ContentType):
        return f"{value.app_label}.{value.model}"
    # For regular related fields, get the primary key
    return  value.pk if value is not None else None

def clean_diff_data(data: dict, exclude_empty_values: bool = True) -> dict:
    """Clean diff data by removing null values."""
    result = {}
    for k, v in data.items():
        if exclude_empty_values:
            if v is None:
                continue
            if isinstance(v, list) and len(v) == 0:
                continue
            if isinstance(v, dict):
                if len(v) == 0:
                    continue
                v = clean_diff_data(v, exclude_empty_values)
            if isinstance(v, str) and v == "":
                continue
        result[k] = v
    return result


def diff_to_change(
    object_type: str,
    prechange_data: dict,
    postchange_data: dict,
    changed_attrs: list[str],
    unresolved_references: list[str],
) -> Change:
    """Convert a diff to a change."""
    change_type = ChangeType.UPDATE if len(prechange_data) > 0 else ChangeType.CREATE
    if change_type == ChangeType.UPDATE and not len(changed_attrs) > 0:
        change_type = ChangeType.NOOP
    primary_value = str(get_primary_value(prechange_data | postchange_data, object_type))
    if primary_value is None:
        primary_value = "(unnamed)"

    prior_id = prechange_data.get("id")
    ref_id = None
    if prior_id is None:
        ref_id = postchange_data.pop("id", None)

    change = Change(
        change_type=change_type,
        before=_tidy(prechange_data),
        data={},
        object_type=object_type,
        object_id=prior_id if isinstance(prior_id, int) else None,
        ref_id=ref_id,
        object_primary_value=primary_value,
        new_refs=unresolved_references,
    )

    if change_type != ChangeType.NOOP:
        change.data = _tidy(postchange_data)

    return change

def _tidy(data: dict) -> dict:
    return sort_dict_recursively(clean_diff_data(data))

def sort_dict_recursively(d):
    """Recursively sorts a dictionary by keys."""
    if isinstance(d, dict):
        return {k: sort_dict_recursively(v) for k, v in sorted(d.items())}
    if isinstance(d, list):
        return [sort_dict_recursively(item) for item in d]
    return d

def generate_changeset(entity: dict, object_type: str) -> ChangeSetResult:
    """Generate a changeset for an entity."""
    try:
        return _generate_changeset(entity, object_type)
    except ChangeSetException:
        raise
    except serializers.ValidationError as e:
        raise error_from_validation_error(e, object_type)
    except Exception as e:
        logger.error(f"Unexpected error generating changeset: {e}")
        raise

def _generate_changeset(entity: dict, object_type: str) -> ChangeSetResult:
    """Generate a changeset for an entity."""
    change_set = ChangeSet()

    warnings = {}
    supported_models = extract_supported_models()
    entities = transform_proto_json(entity, object_type, supported_models)
    by_uuid = {x['_uuid']: x for x in entities}
    for entity in entities:
        prechange_data = {}
        changed_attrs = []
        new_refs = cleanup_unresolved_references(entity)
        object_type = entity.pop("_object_type")
        _ = entity.pop("_uuid")
        instance = entity.pop("_instance", None)
        _merge_warnings(warnings, object_type, entity.pop("_warnings", None))
        if instance:
            # the prior state is another new object...
            if isinstance(instance, str):
                prechange_data = copy.deepcopy(by_uuid[instance])
            # prior state is a model instance
            else:
                prechange_data = prechange_data_from_instance(instance)
                # merge the prior state that we don't want to overwrite with the new state
                # this is also important for custom fields because they do not appear to
                # respsect paritial update serialization.
                entity = _partially_merge(prechange_data, entity, instance)
            changed_data = shallow_compare_dict(
                prechange_data, entity,
            )
            changed_attrs = sorted(changed_data.keys())
        change = diff_to_change(
            object_type,
            prechange_data,
            entity,
            changed_attrs,
            new_refs,
        )

        change_set.changes.append(change)

    has_any_changes = False
    for change in change_set.changes:
        if change.change_type != ChangeType.NOOP:
            has_any_changes = True
            break

    if not has_any_changes:
        change_set.changes = []
    if errors := change_set.validate():
        raise ChangeSetException("Invalid change set", errors)

    if warnings:
        change_set.warnings = warnings

    cs = ChangeSetResult(
        id=change_set.id,
        change_set=change_set,
    )
    return cs

def _partially_merge(prechange_data: dict, postchange_data: dict, instance) -> dict:
    """Merge lists and custom_fields rather than replacing the full value..."""
    result = {}
    for key, value in postchange_data.items():
        # currently we only merge tags, but this could be extended to other reference lists?
        if key == "tags":
            result[key] = _merge_reference_list(prechange_data.get(key, []), value)
        else:
            result[key] = value

    # these are fully merged in from the prechange state because
    # they don't respect partial update serialization.
    if "custom_fields" in postchange_data:
        for key, value in prechange_data.get("custom_fields", {}).items():
            if value is not None and key not in postchange_data["custom_fields"]:
                result["custom_fields"][key] = value
        set_custom_field_defaults(result, instance)
    return result

def _merge_reference_list(prechange_list: list, postchange_list: list) -> list:
    """Merge reference lists rather than replacing the full value."""
    result = set(prechange_list)
    result.update(postchange_list)
    return sort_ints_first(result)

def _merge_warnings(warnings: dict, object_type: str, entity_warnings: dict):
    """Merge warnings for an object type."""
    if not entity_warnings:
        return

    if object_type not in warnings:
        warnings[object_type] = defaultdict(list)
    for key, value in entity_warnings.items():
        warnings[object_type][key] += value



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API Views."""
import logging
import re

from django.apps import apps
from django.db import transaction
from rest_framework import views
from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from netbox_diode_plugin.api.applier import apply_changeset
from netbox_diode_plugin.api.authentication import DiodeOAuth2Authentication
from netbox_diode_plugin.api.common import (
    Change,
    ChangeSet,
    ChangeSetException,
    ChangeSetResult,
)
from netbox_diode_plugin.api.differ import generate_changeset
from netbox_diode_plugin.api.permissions import (
    SCOPE_NETBOX_READ,
    SCOPE_NETBOX_WRITE,
    IsAuthenticated,
    require_scopes,
)

logger = logging.getLogger("netbox.diode_data")

# Try to import Branch model at module level
Branch = None
try:
    if apps.is_installed("netbox_branching"):
        from netbox_branching.models import Branch
except ImportError:
    logger.warning(
        "netbox_branching plugin is installed but models could not be imported"
    )


def get_valid_entity_keys(model_name):
    """
    Get the valid entity keys for a model name.

    This can be snake or lowerCamel case (both are valid for protoJSON)
    """
    s = re.sub(r"([A-Z0-9]{2,})([A-Z])([a-z])", r"\1_\2\3", model_name)
    s = re.sub(r"([a-z])([A-Z])", r"\1_\2", s)
    snake = re.sub(r"_+", "_", s.lower())  # snake
    upperCamel = "".join(
        [word.capitalize() for word in snake.split("_")]
    )  # upperCamelCase
    lowerCamel = upperCamel[0].lower() + upperCamel[1:]  # lowerCamelCase

    return (snake, lowerCamel)


class GenerateDiffView(views.APIView):
    """GenerateDiff view."""

    authentication_classes = [DiodeOAuth2Authentication]
    permission_classes = [IsAuthenticated, require_scopes(SCOPE_NETBOX_READ)]

    def post(self, request, *args, **kwargs):
        """Generate diff for entity."""
        try:
            return self._post(request, *args, **kwargs)
        except ChangeSetException as e:
            result = ChangeSetResult(
                errors=e.errors,
            )
            return Response(result.to_dict(), status=result.get_status_code())
        except Exception:
            import traceback
            traceback.print_exc()
            raise

    def _get_branch_schema_id(self, request):
        """Get branch schema ID from request header or settings."""
        branch_schema_id = request.headers.get("X-NetBox-Branch")

        # If no branch specified in header, check for default branch in settings
        if not branch_schema_id and Branch is not None:
            try:
                from netbox_diode_plugin.models import Setting
                settings = Setting.objects.first()
                if settings and settings.branch:
                    branch_schema_id = settings.branch.schema_id
                    logger.debug(
                        f"Using default branch from settings: {settings.branch.name} ({branch_schema_id})"
                    )
            except Exception as e:
                logger.warning(f"Could not retrieve default branch from settings: {e}")

        return branch_schema_id

    def _add_branch_to_result(self, result, branch_schema_id):
        """Add branch information to the result if branch is available."""
        if branch_schema_id and Branch is not None:
            try:
                branch = Branch.objects.get(schema_id=branch_schema_id)
                result.change_set.branch = {"id": branch.schema_id, "name": branch.name}
            except Branch.DoesNotExist:
                sanitized_branch_id = branch_schema_id.replace('\n', '').replace('\r', '')
                logger.warning(f"Branch with ID {sanitized_branch_id} does not exist")

    def _post(self, request, *args, **kwargs):
        entity = request.data.get("entity")
        object_type = request.data.get("object_type")

        if not entity:
            raise ChangeSetException(
                "validation error",
                errors={
                    "request": {
                        "entity": ["entity is required"]
                    }
                }
            )
        if not object_type:
            raise ChangeSetException(
                "validation error",
                errors={
                    "request": {
                        "object_type": ["object_type is required"]
                    }
                }
            )

        app_label, model_name = object_type.split(".")
        try:
            model_class = apps.get_model(app_label, model_name)
        except LookupError:
            raise ChangeSetException(
                "validation error",
                errors={
                    "request": {
                        "object_type": [f"{object_type} is not supported in this version."]
                    }
                }
            )

        for entity_key in get_valid_entity_keys(model_class.__name__):
            original_entity_data = entity.get(entity_key)
            if original_entity_data:
                break

        if original_entity_data is None:
            raise ChangeSetException(
                "validation error",
                errors={
                    "entity": {
                        entity_key: [f"No data found in expected entity key, got: {entity.keys()}"]
                    }
                }
            )

        result = generate_changeset(original_entity_data, object_type)
        branch_schema_id = self._get_branch_schema_id(request)
        self._add_branch_to_result(result, branch_schema_id)

        return Response(result.to_dict(), status=result.get_status_code())


class ApplyChangeSetView(views.APIView):
    """ApplyChangeSet view."""

    authentication_classes = [DiodeOAuth2Authentication]
    permission_classes = [IsAuthenticated, require_scopes(SCOPE_NETBOX_WRITE)]

    def post(self, request, *args, **kwargs):
        """Apply change set for entity."""
        try:
            return self._post(request, *args, **kwargs)
        except Exception:
            import traceback

            traceback.print_exc()
            raise

    def _post(self, request, *args, **kwargs):
        data = request.data.copy()

        changes = []
        if "changes" in data:
            changes = [
                Change(
                    change_type=change.get("change_type"),
                    object_type=change.get("object_type"),
                    object_id=change.get("object_id"),
                    ref_id=change.get("ref_id"),
                    data=change.get("data"),
                    before=change.get("before"),
                    new_refs=change.get("new_refs", []),
                )
                for change in data["changes"]
            ]
        change_set = ChangeSet(
            id=data.get("id"),
            changes=changes,
        )
        try:
            with transaction.atomic():
                result = apply_changeset(change_set, request)
        except ChangeSetException as e:
            logger.error(f"Error applying change set: {e}")
            result = ChangeSetResult(
                id=change_set.id,
                errors=e.errors,
            )

        return Response(result.to_dict(), status=result.get_status_code())


class GetDefaultBranchView(views.APIView):
    """GetDefaultBranch view."""

    authentication_classes = [DiodeOAuth2Authentication]
    permission_classes = [IsAuthenticated, require_scopes(SCOPE_NETBOX_READ)]

    def get(self, request, *args, **kwargs):
        """Get default branch from settings."""
        branch_data = None

        # Check for default branch in settings
        if Branch is not None:
            try:
                from netbox_diode_plugin.models import Setting
                settings = Setting.objects.first()
                if settings and settings.branch:
                    branch_data = {
                        "id": settings.branch.schema_id,
                        "name": settings.branch.name
                    }
                    logger.debug(
                        f"Default branch from settings: {settings.branch.name} ({settings.branch.schema_id})"
                    )
            except Exception as e:
                logger.warning(f"Could not retrieve default branch from settings: {e}")

        return Response({"branch": branch_data})



#!/usr/bin/env python
# Copyright 2025 NetBox Labs Inc
"""Diode NetBox Plugin - API - Common types and utilities."""

import datetime
import decimal
import logging
import uuid
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from zoneinfo import ZoneInfo

import netaddr
from django.apps import apps
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.db.backends.postgresql.psycopg_any import NumericRange
from extras.models import CustomField
from netaddr.eui import EUI
from rest_framework import status

logger = logging.getLogger("netbox.diode_data")

NON_FIELD_ERRORS = "__all__"

@dataclass
class UnresolvedReference:
    """unresolved reference to an object."""

    object_type: str
    uuid: str

    def __str__(self):
        """String representation of the unresolved reference."""
        return f"new_object:{self.object_type}:{self.uuid}"

    def __eq__(self, other):
        """Equality operator."""
        if not isinstance(other, UnresolvedReference):
            return False
        return self.object_type == other.object_type and self.uuid == other.uuid

    def __hash__(self):
        """Hash function."""
        return hash((self.object_type, self.uuid))

    def __lt__(self, other):
        """Less than operator."""
        if not isinstance(other, UnresolvedReference):
            return False
        return self.object_type < other.object_type or (self.object_type == other.object_type and self.uuid < other.uuid)


class ChangeType(Enum):
    """Change type enum."""

    CREATE = "create"
    UPDATE = "update"
    NOOP = "noop"


@dataclass
class Change:
    """A change to a model instance."""

    change_type: ChangeType
    object_type: str
    object_id: int | None = field(default=None)
    object_primary_value: str | None = field(default=None)
    ref_id: str | None = field(default=None)
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    before: dict | None = field(default=None)
    data: dict | None = field(default=None)
    new_refs: list[str] = field(default_factory=list)

    def to_dict(self) -> dict:
        """Convert the change to a dictionary."""
        return {
            "id": self.id,
            "change_type": self.change_type.value,
            "object_type": self.object_type,
            "object_id": self.object_id,
            "ref_id": self.ref_id,
            "object_primary_value": self.object_primary_value,
            "before": self.before,
            "data": self.data,
            "new_refs": self.new_refs,
        }


@dataclass
class ChangeSet:
    """A set of changes to a model instance."""

    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    changes: list[Change] = field(default_factory=list)
    branch: dict[str, str] | None = field(default=None)  # {"id": str, "name": str}
    warnings: dict | None = field(default=None)

    def to_dict(self) -> dict:
        """Convert the change set to a dictionary."""
        d = {
            "id": self.id,
            "changes": [change.to_dict() for change in self.changes],
            "branch": self.branch,
        }
        if self.warnings:
            d["warnings"] = self.warnings
        return d

    def validate(self) -> dict[str, list[str]]:
        """Validate basics of the change set data."""
        errors = defaultdict(dict)

        for change in self.changes:
            model = apps.get_model(change.object_type)

            change_data = change.data.copy()
            if change.before:
                change_data.update(change.before)

            excluded_relation_fields, rel_errors = self._validate_relations(change_data, model)
            if rel_errors:
                errors[change.object_type] = rel_errors

            try:
                custom_fields = change_data.pop('custom_fields', None)
                if custom_fields:
                    self._validate_custom_fields(custom_fields, model)

                instance = model(**change_data)
                instance.clean_fields(exclude=excluded_relation_fields)
            except ValidationError as e:
                errors[change.object_type].update(_error_dict(e))

        return errors or None

    def _validate_custom_fields(self, data: dict, model: models.Model) -> None:
        custom_fields = {
            cf.name: cf for cf in CustomField.objects.get_for_model(model)
        }

        unknown_errors = []
        for field_name, value in data.items():
            if field_name not in custom_fields:
                unknown_errors.append(f"Unknown field name '{field_name}' in custom field data.")
                continue
        if unknown_errors:
            raise ValidationError({
                "custom_fields": unknown_errors
            })

        req_errors = []
        for field_name, cf in custom_fields.items():
            if cf.required and field_name not in data:
                req_errors.append(f"Custom field '{field_name}' is required.")
        if req_errors:
            raise ValidationError({
                "custom_fields": req_errors
            })

    def _validate_relations(self, change_data: dict, model: models.Model) -> tuple[list[str], dict]:
        # check that there is some value for every required
        # reference field, but don't validate the actual reference.
        # the fields are removed from the change_data so that other
        # fields can be validated by instantiating the model.
        excluded_relation_fields = []
        rel_errors = defaultdict(list)
        for f in model._meta.get_fields():
            if isinstance(f, GenericRelation | GenericForeignKey):
                excluded_relation_fields.append(f.name)
                continue
            if not f.is_relation:
                continue
            field_name = f.name
            excluded_relation_fields.append(field_name)

            if hasattr(f, "related_model") and f.related_model == ContentType:
                change_data.pop(field_name, None)
                base_field = field_name[:-5]
                excluded_relation_fields.append(base_field + "_id")
                value = change_data.pop(base_field + "_id", None)
            else:
                value = change_data.pop(field_name, None)

            if not f.null and not f.blank and not f.many_to_many:
                # this field is a required relation...
                if value is None:
                    rel_errors[f.name].append(f"Field {f.name} is required")
        return excluded_relation_fields, rel_errors


@dataclass
class ChangeSetResult:
    """A result of applying a change set."""

    id: str | None = field(default_factory=lambda: str(uuid.uuid4()))
    change_set: ChangeSet | None = field(default=None)
    errors: dict | None = field(default=None)

    def to_dict(self) -> dict:
        """Convert the result to a dictionary."""
        result = {
            "id": self.id,
            "errors": self.errors,
        }

        if self.change_set:
            result["change_set"] = self.change_set.to_dict()

        return result

    def get_status_code(self) -> int:
        """Get the status code for the result."""
        return status.HTTP_200_OK if not self.errors else status.HTTP_400_BAD_REQUEST


class ChangeSetException(Exception):
    """ChangeSetException is raised when an error occurs while generating or applying a change set."""

    def __init__(self, message, errors=None):
        """Initialize the exception."""
        super().__init__(message)
        self.message = message
        self.errors = errors or {}

    def __str__(self):
        """Return the string representation of the exception."""
        if self.errors:
            return f"{self.message}: {self.errors}"
        return self.message

def _error_dict(e: ValidationError) -> dict:
    """Convert a ValidationError to a dictionary."""
    if hasattr(e, "error_dict"):
        return e.error_dict
    return {
        "__all__": e.error_list
    }

@dataclass
class AutoSlug:
    """A class that marks an auto-generated slug."""

    field_name: str
    value: str

def error_from_validation_error(e, object_name):
    """Convert a from DRF ValidationError to a ChangeSetException."""
    errors = {}
    if e.detail:
        if isinstance(e.detail, dict):
            errors[object_name] = e.detail
        elif isinstance(e.detail, list | tuple):
            errors[object_name] = {
                NON_FIELD_ERRORS: e.detail
            }
        else:
            errors[object_name] = {
                NON_FIELD_ERRORS: [e.detail]
            }
    return ChangeSetException("validation error", errors=errors)

def harmonize_formats(data):
    """Puts all data in a format that can be serialized and compared."""
    match data:
        case None:
            return None
        case str() | int() | float() | bool() | decimal.Decimal() | UnresolvedReference():
            return data
        case dict():
            return {k: harmonize_formats(v) if not k.startswith("_") else v for k, v in data.items()}
        case list() | tuple():
            return [harmonize_formats(v) for v in data]
        case datetime.datetime():
            return data.strftime("%Y-%m-%dT%H:%M:%SZ")
        case datetime.date():
            return data.strftime("%Y-%m-%d")
        case NumericRange():
            return [data.lower, data.upper-1]
        case netaddr.IPNetwork() | EUI() | ZoneInfo():
            return str(data)
        case _:
            logger.warning(f"Unknown type in harmonize_formats: {type(data)}")
            return data

def sort_ints_first(data):
  """Sort a mixed list of ints and other types, putting ints first."""
  return sorted(data, key=lambda x: (not isinstance(x, int), x))



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Database migrations."""

from django.db import migrations, models

import netbox_diode_plugin.models


class Migration(migrations.Migration):
    """Create ClientCredentials model and alter Setting.diode_target field."""

    dependencies = [
        ("netbox_diode_plugin", "0001_squashed_0005"),
    ]

    operations = [
        migrations.CreateModel(
            name="ClientCredentials",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True, primary_key=True, serialize=False
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("view_clientcredentials", "Can view Client Credentials"),
                    (
                        "add_clientcredentials",
                        "Can perform actions on Client Credentials",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.AlterField(
            model_name="setting",
            name="diode_target",
            field=models.CharField(
                max_length=255,
                validators=[netbox_diode_plugin.models.diode_target_validator],
            ),
        ),
    ]



#!/usr/bin/env python
# Copyright 2025 NetBox Labs Inc
"""Diode NetBox Plugin - API - Compatibility Transformations."""

import logging
import re
from collections import defaultdict
from functools import cache

from django.conf import settings
from packaging import version
from utilities.release import load_release_data

logger = logging.getLogger(__name__)

_MIGRATIONS_BY_OBJECT_TYPE = defaultdict(list)

def apply_entity_migrations(data: dict, object_type: str):
    """
    Applies migrations to diode entity data prior to diffing to improve compatibility with current NetBox version.

    These represent cases like deprecated fields that have been replaced with new fields, but
    are supported for backwards compatibility.
    """
    for migration in _MIGRATIONS_BY_OBJECT_TYPE.get(object_type, []):
        logger.debug(f"Applying migration {migration.__name__} for {object_type}")
        migration(data)

def _register_migration(func, min_version, max_version, object_type):
    """Registers a migration function."""
    if in_version_range(min_version, max_version):
        logger.debug(f"Registering migration {func.__name__} for {object_type}.")
        _MIGRATIONS_BY_OBJECT_TYPE[object_type].append(func)
    else:
        logger.debug(f"Skipping migration {func.__name__} for {object_type}: {min_version} to {max_version}.")

@cache
def _current_netbox_version():
    """Returns the current version of NetBox."""
    try:
        return version.parse(settings.RELEASE.version)
    except Exception:
        logger.exception("Failed to determine current version of NetBox.")
        return (0, 0, 0)

def in_version_range(min_version: str | None, max_version: str | None):
    """Returns True if the current version of NetBox is within the given version range."""
    min_version = version.parse(min_version) if min_version else None
    max_version = version.parse(max_version) if max_version else None
    current_version = _current_netbox_version()
    if min_version and current_version < min_version:
        return False
    if max_version and current_version > max_version:
        return False
    return True

def diode_migration(min_version: str, max_version: str | None, object_type: str):
    """Decorator to mark a function as a diode migration."""
    def decorator(func):
        _register_migration(func, min_version, max_version, object_type)
        return func
    return decorator

@diode_migration(min_version="4.3.0", max_version=None, object_type="ipam.service")
def _migrate_service_parent_object(data: dict):
    """Transforms ipam.service device and virtual_machine references to parent_object."""
    device = data.pop("device", None)
    if device:
        if data.get("parent_object_device") is None:
            data["parent_object_device"] = device
        # else ignored.

    virtual_machine = data.pop("virtual_machine", None)
    if virtual_machine:
        if data.get("parent_object_virtual_machine") is None:
            data["parent_object_virtual_machine"] = virtual_machine
        # else ignored.

@diode_migration(min_version="4.3.0", max_version=None, object_type="tenancy.contact")
def _migrate_contact_group(data: dict):
    """Transforms tenancy.contact group references to groups."""
    group = data.pop("group", None)
    if group:
        if data.get("groups") is None:
            data["groups"] = [group]
        # else ignored.

@diode_migration(min_version="4.2.0", max_version="4.2.99", object_type="ipam.service")
def _migrate_service_parent_object_down(data: dict):
    """Transforms ipam.service parent_object to device and virtual_machine."""
    parent_object_vm = data.pop("parent_object_virtual_machine", None)
    if parent_object_vm and data.get("virtual_machine") is None:
        data["virtual_machine"] = parent_object_vm
    parent_object_device = data.pop("parent_object_device", None)
    if parent_object_device and data.get("device") is None:
        data["device"] = parent_object_device

@diode_migration(min_version="4.2.0", max_version="4.2.99", object_type="tenancy.contact")
def _migrate_contact_group_down(data: dict):
    """Transforms tenancy.contact groups to group."""
    groups = data.pop("groups", None)
    if groups and len(groups) == 1:
        data["group"] = groups[0]



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API - Object resolution for diffing."""

import copy
import datetime
import graphlib
import json
import logging
import re
from collections import defaultdict
from functools import lru_cache
from uuid import uuid4

from django.utils.text import slugify
from extras.models.customfields import CustomField
from rest_framework import serializers

from .common import NON_FIELD_ERRORS, AutoSlug, ChangeSetException, UnresolvedReference, harmonize_formats, sort_ints_first
from .compat import apply_entity_migrations
from .matcher import find_existing_object, fingerprints
from .plugin_utils import (
    CUSTOM_FIELD_OBJECT_REFERENCE_TYPE,
    apply_format_transformations,
    get_json_ref_info,
    get_primary_value,
    legal_fields,
)

logger = logging.getLogger("netbox.diode_data")

@lru_cache(maxsize=128)
def _camel_to_snake_case(name):
    """Convert camelCase string to snake_case."""
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()

# these are implied values pushed down to referenced objects.
_NESTED_CONTEXT = {
    "dcim.interface": {
        # interface.primary_mac_address -> mac_address.assigned_object = interface
        "primary_mac_address": lambda object_type, uuid: {
            "assigned_object_type": object_type,
            "assigned_object_id": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
    "virtualization.vminterface": {
        # interface.primary_mac_address -> mac_address.assigned_object = vinterface
        "primary_mac_address": lambda object_type, uuid: {
            "assigned_object_type": object_type,
            "assigned_object_id": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
    "virtualization.virtualmachine": {
        "primary_ip4": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
        "primary_ip6": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
    "dcim.virtualdevicecontext": {
        "primary_ip4": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
        "primary_ip6": lambda object_type, uuid: {
            "__force_after": UnresolvedReference(object_type=object_type, uuid=uuid),
        },
    },
}

def _no_context(object_type, uuid):
    return None

def _nested_context(object_type, uuid, field_name):
    return _NESTED_CONTEXT.get(object_type, {}).get(field_name, _no_context)(object_type, uuid)

_IS_CIRCULAR_REFERENCE = {
    "dcim.interface": frozenset(["primary_mac_address"]),
    "virtualization.vminterface": frozenset(["primary_mac_address"]),
    "dcim.device": frozenset(["primary_ip4", "primary_ip6", "oob_ip"]),
    "dcim.virtualdevicecontext": frozenset(["primary_ip4", "primary_ip6"]),
    "virtualization.virtualmachine": frozenset(["primary_ip4", "primary_ip6"]),
    "circuits.provider": frozenset(["accounts"]),
    "dcim.modulebay": frozenset(["module"]), # this isn't  allowed to be circular, but gives a better error
}

def _is_circular_reference(object_type, field_name):
    return field_name in _IS_CIRCULAR_REFERENCE.get(object_type, frozenset())

def transform_proto_json(proto_json: dict, object_type: str, supported_models: dict) -> list[dict]: # noqa: C901
    """
    Transform keys of proto json dict to flattened dictionaries with model field keys.

    This also handles placing `_type` fields for generic references,
    a certain form of deduplication and resolution of existing objects.
    """
    entities = _transform_proto_json_1(proto_json, object_type, supported_models)

    entities = _topo_sort(entities)
    deduplicated = _fingerprint_dedupe(entities)
    deduplicated = _topo_sort(deduplicated)
    _set_auto_slugs(deduplicated, supported_models)
    _handle_cached_scope(deduplicated, supported_models)
    resolved = _resolve_existing_references(deduplicated)
    _strip_cached_scope(resolved)
    defaulted = _set_defaults(resolved, supported_models)

    # handle post-create steps
    output = _handle_post_creates(defaulted)

    _check_unresolved_refs(output)
    for entity in output:
        entity.pop('_refs', None)

    return output

def _transform_proto_json_1(proto_json: dict, object_type: str, supported_models: dict, context=None) -> list[dict]: # noqa: C901
    uuid = str(uuid4())
    node = {
        "_object_type": object_type,
        "_uuid": uuid,
        "_refs": set(),
        "_warnings": {},
    }

    # handle camelCase protoJSON if provided...
    proto_json = _ensure_snake_case(proto_json, object_type)
    apply_format_transformations(proto_json, object_type)
    apply_entity_migrations(proto_json, object_type)

    # context pushed down from parent nodes
    if context is not None:
        for k, v in context.items():
            if not k.startswith("_"):
                node[k] = v
            if isinstance(v, UnresolvedReference):
                node['_refs'].add(v.uuid)

    nodes = [node]
    post_create = None

    # special handling for custom fields
    custom_fields = dict.pop(proto_json, "custom_fields", {})
    if custom_fields:
        custom_fields, custom_fields_refs, nested = _prepare_custom_fields(object_type, custom_fields, supported_models)
        node['custom_fields'] = custom_fields
        node['_refs'].update(custom_fields_refs)
        nodes += nested

    supported_fields = _supported_diode_fields(object_type, supported_models)
    def is_supported(field_name, ref_info):
        if ref_info is None:
            return field_name in supported_fields
        if ref_info.object_type not in supported_models:
            return False
        if ref_info.is_generic:
            return ref_info.field_name + "_type" in supported_fields
        return ref_info.field_name in supported_fields

    for key, value in proto_json.items():
        ref_info = get_json_ref_info(object_type, key)
        if not is_supported(key, ref_info):
            node['_warnings'][key] = ["Ignored unsupported field."]
            continue

        if ref_info is None:
            node[key] = copy.deepcopy(value)
            continue

        nested_context = _nested_context(object_type, uuid, ref_info.field_name)
        field_name = ref_info.field_name
        is_circular = _is_circular_reference(object_type, field_name)

        if ref_info.is_generic:
            node[field_name + "_type"] = ref_info.object_type
            field_name = field_name + "_id"

        refs = []
        ref_value = None
        if isinstance(value, list):
            ref_value = []
            for item in value:
                nested = _transform_proto_json_1(item, ref_info.object_type, supported_models, nested_context)
                nodes += nested
                ref_uuid = nested[0]['_uuid']
                ref_value.append(UnresolvedReference(
                    object_type=ref_info.object_type,
                    uuid=ref_uuid,
                ))
                refs.append(ref_uuid)
        else:
            nested = _transform_proto_json_1(value, ref_info.object_type, supported_models, nested_context)
            nodes += nested
            ref_uuid = nested[0]['_uuid']
            ref_value = UnresolvedReference(
                object_type=ref_info.object_type,
                uuid=ref_uuid,
            )
            refs.append(ref_uuid)

        if is_circular:
            if post_create is None:
                post_create = {
                    "_uuid": str(uuid4()),
                    "_object_type": object_type,
                    "_refs": set(),
                    "_instance": node['_uuid'],
                    "_is_post_create": True,
                }
            post_create[field_name] = ref_value
            post_create['_refs'].update(refs)
            post_create['_refs'].add(node['_uuid'])
            continue

        node[field_name] = ref_value
        node['_refs'].update(refs)

    if post_create:
        nodes.append(post_create)

    return nodes

def _ensure_snake_case(proto_json: dict, object_type: str) -> dict:
    fields = legal_fields(object_type)
    out = {}
    for k, v in proto_json.items():
        if k in fields or get_json_ref_info(object_type, k):
            out[k] = v
            continue
        snake_key = _camel_to_snake_case(k)
        if snake_key in fields or get_json_ref_info(object_type, snake_key):
            out[snake_key] = v
        else:
            # error?
            sanitized_k = k.replace('\n', '').replace('\r', '')
            sanitized_snake_key = snake_key.replace('\n', '').replace('\r', '')
            sanitized_object_type = object_type.replace('\n', '').replace('\r', '')
            logger.warning(f"Unknown field {sanitized_k}/{sanitized_snake_key} is not legal for {sanitized_object_type}, skipping...")
    return out


def _topo_sort(entities: list[dict]) -> list[dict]:
    """Topologically sort entities by reference."""
    by_uuid = {e['_uuid']: e for e in entities}
    graph = defaultdict(set)
    for entity in entities:
        graph[entity['_uuid']] = entity['_refs'].copy()

    try:
        ts = graphlib.TopologicalSorter(graph)
        order = tuple(ts.static_order())
        return [by_uuid[uuid] for uuid in order]
    except graphlib.CycleError as e:
        # TODO the cycle error references the cycle here ...
        raise ChangeSetException(f"Circular reference in entities: {e}", errors={
            NON_FIELD_ERRORS: {
                NON_FIELD_ERRORS: "Unable to resolve circular reference in entities",
            }
        })


def _set_defaults(entities: list[dict], supported_models: dict):
    out = []
    for entity in entities:
        entity = copy.deepcopy(entity)
        model_fields = supported_models.get(entity['_object_type'])
        if model_fields is None:
            raise serializers.ValidationError({
                NON_FIELD_ERRORS: [f"Model for object type {entity['_object_type']} is not supported"]
            })

        auto_slug = entity.pop("_auto_slug", None)
        if entity.get("_instance"):
            out.append(entity)
            continue

        if auto_slug:
            if auto_slug.field_name not in entity:
                entity[auto_slug.field_name] = auto_slug.value

        legal = legal_fields(entity['_object_type'])
        for field_name, field_info in model_fields.get('fields', {}).items():
            if field_name not in legal:
                continue
            if entity.get(field_name) is None and field_info.get("default") is not None:
                default = field_info["default"]
                if callable(default):
                    default = default()
                entity[field_name] = default
        set_custom_field_defaults(entity, model_fields['model'])
        out.append(harmonize_formats(entity))
    return out

def _handle_cached_scope(entities: list[dict], supported_models: dict):
    by_type_id = {
        (entity['_object_type'], entity['_uuid']): entity
        for entity in entities
    }
    for entity in entities:
        model = supported_models.get(entity['_object_type'], {}).get("model")
        if _has_cached_scope(model):
            _handle_cached_scope_1(entity, by_type_id)

def _strip_cached_scope(entities: list[dict]):
    for entity in entities:
        entity.pop("_region", None)
        entity.pop("_site_group", None)
        entity.pop("_site", None)
        entity.pop("_location", None)

@lru_cache(maxsize=256)
def _has_cached_scope(model):
    return  hasattr(model, "cache_related_objects") and hasattr(model, "scope")

def _handle_cached_scope_1(entity: dict, by_type_id: dict):
    # these are some auto-set fields that cache scope information,
    # some indexes rely on them. Here we attempt to emulate that behavior
    # for the purpose of matching.  These generally only exist after save.
    scope_type = entity.get("scope_type")
    scope_id = entity.get("scope_id")

    if scope_type and scope_id:
        scope = by_type_id.get((scope_type, scope_id.uuid))
        if scope_type == "dcim.region":
            _cache_region_ref(entity, scope_id)
        elif scope_type == "dcim.sitegroup":
            _cache_site_group_ref(entity, scope_id)
        elif scope_type == "dcim.site":
            _cache_site_ref(entity, scope_id)
            _cache_region_ref(entity, scope.get("region"))
            _cache_site_group_ref(entity, scope.get("group"))
        elif scope_type == "dcim.location":
            _cache_location_ref(entity, scope_id)
            site_ref = scope.get("site")
            if site_ref is not None and isinstance(site_ref, UnresolvedReference):
                _cache_site_ref(entity, site_ref)
                site_obj = by_type_id.get((site_ref.object_type, site_ref.uuid))
                if site_obj is not None:
                    _cache_region_ref(entity, site_obj.get("region"))
                    _cache_site_group_ref(entity, site_obj.get("group"))

def _cache_region_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_region"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def _cache_site_group_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_site_group"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def _cache_site_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_site"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def _cache_location_ref(entity: dict, ref: UnresolvedReference|None):
    if ref is None:
        return
    entity["_location"] = UnresolvedReference(
        object_type=ref.object_type,
        uuid=ref.uuid,
    )

def set_custom_field_defaults(entity: dict, model):
    """Set default values for custom fields in an entity."""
    custom_fields = CustomField.objects.get_for_model(model)
    if custom_fields:
        custom_field_data = entity.get('custom_fields')
        if custom_field_data is None:
            custom_field_data = {}
            entity['custom_fields'] = custom_field_data
        for cf in custom_fields:
            if cf.name not in custom_field_data or custom_field_data[cf.name] is None:
                custom_field_data[cf.name] = cf.default

def _set_auto_slugs(entities: list[dict], supported_models: dict):
    for entity in entities:
        model_fields = supported_models.get(entity['_object_type'])
        if model_fields is None:
            raise serializers.ValidationError({
                NON_FIELD_ERRORS: [f"Model for object type {entity['_object_type']} is not supported"]
            })

        for field_name, field_info in model_fields.get('fields', {}).items():
            if field_info["type"] == "SlugField" and entity.get(field_name) is None:
                slug = _generate_slug(entity['_object_type'], entity)
                if slug is not None:
                    # this is provisionally set but will not be used
                    # if the entity is identified by other means...
                    entity['_auto_slug'] = AutoSlug(field_name=field_name, value=slug)

def _generate_slug(object_type, data):
    """Generate a slug for a model instance."""
    source_value = get_primary_value(data, object_type)
    if source_value is not None:
        return slugify(str(source_value))
    return None

def _fingerprint_dedupe(entities: list[dict]) -> list[dict]: # noqa: C901
    """
    Deduplicates/merges entities by fingerprint.

    *list must be in topo order by reference already*
    """
    by_uuid = {}
    by_fp = {}
    deduplicated = []
    new_refs = {} # uuid -> uuid

    for entity in entities:
        if entity.get('_is_post_create'):
            fp = entity['_uuid']
            existing_uuid = None
        else:
            _update_unresolved_refs(entity, new_refs)
            fps = fingerprints(entity, entity['_object_type'])
            for fp in fps:
                existing_uuid = by_fp.get(fp)
                if existing_uuid is not None:
                    break

        if existing_uuid is None:
            new_entity = copy.deepcopy(entity)
            _update_unresolved_refs(new_entity, new_refs)
            primary_uuid = new_entity['_uuid']
            for fp in fps:
                by_fp[fp] = primary_uuid
            by_uuid[primary_uuid] = new_entity
            deduplicated.append(primary_uuid)
        else:
            existing = by_uuid[existing_uuid]
            new_refs[entity['_uuid']] = existing['_uuid']
            merged = _merge_nodes(existing, entity)
            _update_unresolved_refs(merged, new_refs)
            for fp in fps:
                by_fp[fp] = existing_uuid
            by_uuid[existing_uuid] = merged
            deduplicated.append(existing_uuid)

    return [by_uuid[u] for u in deduplicated]

def _merge_nodes(a: dict, b: dict) -> dict:
    """
    Merges two nodes.

    If there are any conflicts, an error is raised.
    Ignores conflicts in fields that start with an underscore,
    preferring a's value.
    """
    merged = copy.deepcopy(a)
    merged['_refs'] = a['_refs'] | b['_refs']

    for k, v in b.items():
        if k.startswith("_"):
            continue
        if k in merged and merged[k] != v:
            ov = {
                ok: v for ok, v in a.items()
                if ok != k and not ok.startswith("_")
            }
            raise serializers.ValidationError({
                NON_FIELD_ERRORS: [
                    f"Conflicting values for '{k}' merging duplicate {a.get('_object_type')},"
                    f" `{merged[k]}` != `{v}` other values : {ov}"]
            })
        merged[k] = v
    return merged


def _update_unresolved_refs(entity, new_refs):
    if entity.get('_is_post_create'):
        instance_uuid = entity['_instance']
        entity['_instance'] = new_refs.get(instance_uuid, instance_uuid)

    entity['_refs'] = {new_refs.get(r,r) for r in entity['_refs']}
    _update_dict_refs(entity, new_refs)


def _update_dict_refs(data, new_refs):
    for k, v in data.items():
        if isinstance(v, UnresolvedReference) and v.uuid in new_refs:
            v.uuid = new_refs[v.uuid]
        elif isinstance(v, list | tuple):
            for item in v:
                if isinstance(item, UnresolvedReference) and item.uuid in new_refs:
                    item.uuid = new_refs[item.uuid]
        elif isinstance(v, dict):
            _update_dict_refs(v, new_refs)


def _resolve_existing_references(entities: list[dict]) -> list[dict]:
    seen = {}
    new_refs = {}
    resolved = []

    for data in entities:
        object_type = data['_object_type']
        data = copy.deepcopy(data)
        _update_resolved_refs(data, new_refs)

        if data.get('_is_post_create'):
            resolved.append(data)
            continue

        existing = find_existing_object(data, object_type)
        if existing is not None:
            fp = (object_type, existing.id)
            if fp in seen:
                logger.warning(f"objects resolved to the same existing id after deduplication: {seen[fp]} and {data}")
            else:
                seen[fp] = data
            data['id'] = existing.id
            data['_instance'] = existing
            new_refs[data['_uuid']] = existing.id
            resolved.append(data)
        else:
            data['id'] = UnresolvedReference(object_type, data['_uuid'])
            _update_resolved_refs(data, new_refs)
            resolved.append(data)
    return resolved

def _update_resolved_refs(data, new_refs):
    for k, v in list(data.items()):
        if isinstance(v, UnresolvedReference) and v.uuid in new_refs:
            data[k] = new_refs[v.uuid]
        elif isinstance(v, list | tuple):
            new_items = []
            has_refs = False
            for item in v:
                if isinstance(item, UnresolvedReference) and item.uuid in new_refs:
                    new_items.append(new_refs[item.uuid])
                    has_refs = True
                else:
                    new_items.append(item)
            if has_refs:
                data[k] = sort_ints_first(new_items)
        elif isinstance(v, dict):
            _update_resolved_refs(v, new_refs)

def cleanup_unresolved_references(data: dict) -> list[str]:
    """Find and stringify unresolved references in fields."""
    unresolved = set()
    for k, v in data.items():
        if isinstance(v, UnresolvedReference):
            if k != 'id':
                unresolved.add(k)
            data[k] = str(v)
        elif isinstance(v, list | tuple):
            items = []
            for item in v:
                if isinstance(item, UnresolvedReference):
                    unresolved.add(k)
                    items.append(str(item))
                else:
                    items.append(item)
            data[k] = items
        elif isinstance(v, dict):
            for uu in cleanup_unresolved_references(v):
                unresolved.add(f"{k}.{uu}")
    return sorted(unresolved)

def _handle_post_creates(entities: list[dict]) -> list[str]:
    """Merges any unnecessary post-create steps for existing objects."""
    by_uuid = {e['_uuid']: (i, e) for i, e in enumerate(entities)}
    out = []
    for entity in entities:
        is_post_create = entity.pop('_is_post_create', False)
        if not is_post_create:
            out.append(entity)
            continue

        instance = entity.get('_instance')
        prior_index, prior_entity = by_uuid[instance]

        # a post create can be merged whenever the entities it relies on
        # already exist (were resolved) or there are no dependencies between
        # the object being updated and the post-create.
        can_merge = all(
            by_uuid[r][1].get('_instance') is not None
            for r in entity['_refs']
        ) or sorted(by_uuid[r][0] for r in entity['_refs'])[-1] == prior_index

        if can_merge:
            prior_entity.update([x for x in entity.items() if not x[0].startswith('_')])
        else:
            entity['id'] = prior_entity['id']
            out.append(entity)

    return out

def _check_unresolved_refs(entities: list[dict]) -> list[str]:
    seen = set()
    for e in entities:
        seen.add((e['_object_type'], e['_uuid']))
        for k, v in e.items():
            if isinstance(v, UnresolvedReference):
                if (v.object_type, v.uuid) not in seen:
                    raise ChangeSetException(
                        f"Unresolved reference {v} in {e} does not refer to a prior created object (circular reference?)",
                        errors={
                            e['_object_type']: {
                                k: ["unable to resolve reference"],
                            }
                        }
                    )


def _prepare_custom_fields(object_type: str, custom_fields: dict, supported_models: dict) -> tuple[dict, set, list]: # noqa: C901
    """Prepare custom fields for transformation."""
    out = {}
    refs = set()
    nodes = []
    for key, value in custom_fields.items():
        keyname = key
        try:
            value_type, value = _pop_custom_field_type_and_value(value)
            if value_type in ("text", "long_text", "decimal", "boolean", "datetime", "selection", "url", "multiple_selection"):
                out[key] = value
            elif value_type == "date":
                # truncate to YYYY-MM-DD
                try:
                    out[key] = datetime.datetime.fromisoformat(value).strftime("%Y-%m-%d")
                except Exception:
                    out[key] = value
            elif value_type == "integer":
                out[key] = int(value)
            elif value_type == "json":
                out[key] = _prepare_custom_json(value)
            elif value_type == "object":
                nested = _prepare_custom_ref(value, supported_models)
                ref = nested[0]
                refs.add(ref['_uuid'])
                nodes += nested
                out[key] = UnresolvedReference(
                    object_type=ref['_object_type'],
                    uuid=ref['_uuid'],
                )
            elif value_type == "multiple_objects":
                vals = []
                for i, item in enumerate(value):
                    keyname = f"{key}[{i}]"
                    nested = _prepare_custom_ref(item, supported_models)
                    ref = nested[0]
                    refs.add(ref['_uuid'])
                    nodes += nested
                    vals.append(UnresolvedReference(
                        object_type=ref['_object_type'],
                        uuid=ref['_uuid'],
                    ))
                out[key] = vals
            else:
                raise serializers.ValidationError({
                    keyname: [f"Custom field {keyname} has unknown type: {value_type}"]
                })
        except ValueError as e:
            raise ChangeSetException(
                f"Custom field {keyname} is invalid: {value}",
                errors={
                    object_type: {keyname: [str(e)]},
                }
            )
    return out, refs, nodes


def _prepare_custom_json(data: dict) -> dict:
    try:
        return json.loads(data)
    except json.JSONDecodeError:
        raise ValueError("failed to parse as JSON")


def _pop_custom_field_type_and_value(data: dict):
    if not isinstance(data, dict) or len(data) != 1:
        raise ValueError("custom field value must be a dictionary with a single key")
    value_type, value = data.popitem()
    return value_type, value


def _prepare_custom_ref(data: dict, supported_models: dict) -> list[dict]:
    if not isinstance(data, dict) or len(data) != 1:
        raise ValueError("must be a dictionary with a single key")

    field_name, value = data.popitem()
    if not isinstance(value, dict):
        raise ValueError(f"{field_name} must be a dictionary")
    ref_info = get_json_ref_info(CUSTOM_FIELD_OBJECT_REFERENCE_TYPE, field_name)
    if ref_info is None:
        raise ValueError(f"{field_name} is not a supported custom field reference type")

    object_type = ref_info.object_type
    return _transform_proto_json_1(value, object_type, supported_models)

def _supported_diode_fields(object_type, supported_models: dict) -> list[str]:
    """
    Get the supported diode fields for a model.

    This excludes fields that are not supported by the current version of NetBox
    that the plugin is installed in. i.e. fields from older or newer versions of
    NetBox that are also supported by the plugin.
    """
    model = supported_models.get(object_type)
    if not model:
        raise serializers.ValidationError({
            NON_FIELD_ERRORS: [f"{object_type} is not supported in this version."]
        })
    model_fields = set(model.get("fields", {}).keys())
    diode_fields = set(legal_fields(object_type))
    return list(model_fields.intersection(diode_fields))



# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Client."""

import logging

from netbox_diode_plugin.diode.clients import get_api_client

logger = logging.getLogger("netbox.diode_data")


def create_client(request, client_name: str, scope: str):
    """Create client."""
    logger.info(f"Creating client {client_name} with scope {scope}")
    return get_api_client().create_client(client_name, scope)


def delete_client(request, client_id: str):
    """Delete client."""
    sanitized_client_id = client_id.replace("\n", "").replace("\r", "")
    logger.info(f"Deleting client {sanitized_client_id}")
    return get_api_client().delete_client(client_id)


def list_clients(request):
    """List clients."""
    logger.info("Listing clients")
    response = get_api_client().list_clients()
    return response["data"]


def get_client(request, client_id: str):
    """Get client."""
    sanitized_client_id = client_id.replace("\n", "").replace("\r", "")
    logger.info(f"Getting client {sanitized_client_id}")
    return get_api_client().get_client(client_id)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs Inc
"""NetBox Diode Data - API supported models."""

import importlib
import logging
import time
from functools import cache, lru_cache

from django.apps import apps
from django.db import models
from django.db.models import ManyToOneRel
from django.db.models.fields import NOT_PROVIDED
from rest_framework import serializers
from utilities.api import get_serializer_for_model as netbox_get_serializer_for_model

from netbox_diode_plugin.api.plugin_utils import legal_fields, legal_object_types

logger = logging.getLogger(__name__)


@cache
def extract_supported_models() -> dict[str, dict]:
    """Extract supported models from installed NetBox apps / version."""
    start_ts = time.time()

    extracted_models: dict[str, dict] = {}
    possible_object_types = legal_object_types()

    for object_type in possible_object_types:
        try:
            app_label, model_name = object_type.split(".")
            model = apps.get_model(app_label, model_name)
        except LookupError:
            continue

        try:
            fields = _get_model_fields(model)
            if not fields:
                continue

            extracted_models[object_type] = {
                "fields": fields,
                "model": model,
            }
        except Exception as e:
            logger.error(f"extract_supported_models: {model.__name__} error: {e}")

    finish_ts = time.time()
    elapsed_millis = (finish_ts - start_ts) * 1000
    logger.info(
        f"done extracting supported diode models in {elapsed_millis:.2f} milliseconds - extracted_models: {len(extracted_models)}"
    )

    return extracted_models

def _get_model_fields(model_class) -> dict:
    """Get the fields for the model."""
    legal = legal_fields(model_class)
    fields_info: dict[str, dict] = {}
    for field in model_class._meta.get_fields():
        field_name = field.name
        if field_name not in legal and field_name != 'id':
            continue

        field_info = {
            "type": field.get_internal_type(),
        }

        # Collect default values
        default_value = None
        if hasattr(field, "default"):
            default_value = (
                field.default if field.default not in (NOT_PROVIDED, dict) else None
            )
        field_info["default"] = default_value
        fields_info[field_name] = field_info

    return fields_info

@lru_cache(maxsize=128)
def get_serializer_for_model(model, prefix=""):
    """Cached wrapper for NetBox's get_serializer_for_model function."""
    return netbox_get_serializer_for_model(model, prefix)



# !/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Plugin Settings."""

import logging
import os
from urllib.parse import urlparse

from django.contrib.auth import get_user_model
from netbox.plugins import get_plugin_config

__all__ = (
    "get_diode_auth_introspect_url",
    "get_diode_user",
)

User = get_user_model()

logger = logging.getLogger("netbox.diode_data")

def _parse_diode_target(target: str) -> tuple[str, str, bool]:
    """Parse the target into authority, path and tls_verify."""
    parsed_target = urlparse(target)

    if parsed_target.scheme not in ["grpc", "grpcs", "http", "https"]:
        raise ValueError("target should start with grpc://, grpcs://, http:// or https://")

    tls_verify = parsed_target.scheme in ["grpcs", "https"]

    authority = parsed_target.netloc

    if ":" not in authority:
        if parsed_target.scheme in ["grpc", "http"]:
            authority += ":80"
        elif parsed_target.scheme in ["grpcs", "https"]:
            authority += ":443"

    return authority, parsed_target.path, tls_verify


def get_diode_auth_introspect_url():
    """Returns the Diode Auth introspect URL."""
    diode_auth_base_url = get_diode_auth_base_url()
    return f"{diode_auth_base_url}/introspect"

def get_diode_auth_base_url():
    """Returns the Diode Auth service base URL."""
    diode_target = get_plugin_config("netbox_diode_plugin", "diode_target")
    diode_target_override = get_plugin_config(
        "netbox_diode_plugin", "diode_target_override"
    )

    authority, path, tls_verify = _parse_diode_target(
        diode_target_override or diode_target
    )
    scheme = "https" if tls_verify else "http"
    path = path.rstrip("/")

    return f"{scheme}://{authority}{path}/auth"

def get_diode_credentials():
    """Returns the Diode credentials."""
    client_id = get_plugin_config("netbox_diode_plugin", "netbox_to_diode_client_id")
    secrets_path = get_plugin_config("netbox_diode_plugin", "secrets_path")
    secret_name = get_plugin_config("netbox_diode_plugin", "netbox_to_diode_client_secret_name")
    client_secret = get_plugin_config("netbox_diode_plugin", "netbox_to_diode_client_secret")

    if not client_secret:
        secret_file = os.path.join(secrets_path, secret_name)
        client_secret = _read_secret(secret_file, client_secret)

    return client_id, client_secret

def get_diode_max_auth_retries():
    """Returns the Diode max auth retries."""
    return get_plugin_config("netbox_diode_plugin", "diode_max_auth_retries")

# Read secret from file
def _read_secret(secret_file: str, default: str | None = None) -> str | None:
    try:
        f = open(secret_file, encoding='utf-8')
    except OSError:
        return default
    else:
        with f:
            return f.readline().strip()

def get_diode_user():
    """Returns the Diode user."""
    diode_username = get_plugin_config("netbox_diode_plugin", "diode_username")

    try:
        diode_user = User.objects.get(username=diode_username)
    except User.DoesNotExist:
        diode_user = User.objects.create(username=diode_username, is_active=True)

    return diode_user

def get_required_token_audience():
    """Returns the require token audience."""
    return get_plugin_config("netbox_diode_plugin", "required_token_audience")



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API - Object matching utilities."""

import logging
from dataclasses import dataclass
from functools import cache, lru_cache

import netaddr
from django.contrib.contenttypes.fields import ContentType
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import F, Value
from django.db.models.fields import SlugField
from django.db.models.lookups import Exact
from django.db.models.query_utils import Q
from extras.models.customfields import CustomField

from .common import UnresolvedReference
from .compat import in_version_range
from .plugin_utils import content_type_id, get_object_type, get_object_type_model

logger = logging.getLogger(__name__)

#
# these matchers are not driven by netbox unique constraints,
# but are logical criteria that may be used to match objects.
# These should represent the likely intent of a user when
# matching existing objects.
#
_LOGICAL_MATCHERS = {
    "dcim.macaddress": lambda: [
        ObjectMatchCriteria(
            fields=("mac_address", "assigned_object_type", "assigned_object_id"),
            name="logical_mac_address_within_parent",
            model_class=get_object_type_model("dcim.macaddress"),
            condition=Q(assigned_object_id__isnull=False),
        ),
        ObjectMatchCriteria(
            fields=("mac_address", "assigned_object_type", "assigned_object_id"),
            name="logical_mac_address_within_parent",
            model_class=get_object_type_model("dcim.macaddress"),
            condition=Q(assigned_object_id__isnull=True),
        ),
    ],
    "ipam.aggregate": lambda: [
        ObjectMatchCriteria(
            fields=("prefix",),
            name="logical_aggregate_prefix_no_rir",
            model_class=get_object_type_model("ipam.aggregate"),
            condition=Q(rir__isnull=True),
        ),
        ObjectMatchCriteria(
            fields=("prefix", "rir"),
            name="logical_aggregate_prefix_within_rir",
            model_class=get_object_type_model("ipam.aggregate"),
            condition=Q(rir__isnull=False),
        ),
    ],
    "ipam.ipaddress": lambda: [
        GlobalIPNetworkIPMatcher(
            ip_fields=("address",),
            vrf_field="vrf",
            model_class=get_object_type_model("ipam.ipaddress"),
            name="logical_ip_address_global_no_vrf",
        ),
        VRFIPNetworkIPMatcher(
            ip_fields=("address",),
            vrf_field="vrf",
            model_class=get_object_type_model("ipam.ipaddress"),
            name="logical_ip_address_within_vrf",
        ),
    ],
    "ipam.iprange": lambda: [
        GlobalIPNetworkIPMatcher(
            ip_fields=("start_address", "end_address"),
            vrf_field="vrf",
            model_class=get_object_type_model("ipam.iprange"),
            name="logical_ip_range_start_end_global_no_vrf",
        ),
        VRFIPNetworkIPMatcher(
            ip_fields=("start_address", "end_address"),
            vrf_field="vrf",
            model_class=get_object_type_model("ipam.iprange"),
            name="logical_ip_range_start_end_within_vrf",
        ),
    ],
    "ipam.prefix": lambda: [
         ObjectMatchCriteria(
            fields=("prefix",),
            name="logical_prefix_global_no_vrf",
            model_class=get_object_type_model("ipam.prefix"),
            condition=Q(vrf__isnull=True),
        ),
        ObjectMatchCriteria(
            fields=("prefix", "vrf"),
            name="logical_prefix_within_vrf",
            model_class=get_object_type_model("ipam.prefix"),
            condition=Q(vrf__isnull=False),
        ),
    ],
    "virtualization.cluster": lambda: [
        ObjectMatchCriteria(
            fields=("name", "scope_type", "scope_id"),
            name="logical_cluster_within_scope",
            model_class=get_object_type_model("virtualization.cluster"),
            condition=Q(scope_type__isnull=False),
        ),
        ObjectMatchCriteria(
            fields=("name",),
            name="logical_cluster_with_no_scope_or_group",
            model_class=get_object_type_model("virtualization.cluster"),
            condition=Q(scope_type__isnull=True, group__isnull=True),
        ),
    ],
    "ipam.vlan": lambda: [
        ObjectMatchCriteria(
            fields=("vid",),
            name="logical_vlan_vid_no_group_or_svlan_or_site",
            model_class=get_object_type_model("ipam.vlan"),
            condition=Q(group__isnull=True, qinq_svlan__isnull=True, site__isnull=True),
        ),
        ObjectMatchCriteria(
            fields=("vid", "site"),
            name="logical_vlan_in_site",
            model_class=get_object_type_model("ipam.vlan"),
            condition=Q(group__isnull=True, qinq_svlan__isnull=True, site__isnull=False),
        ),
    ],
    "ipam.vlangroup": lambda: [
        ObjectMatchCriteria(
            fields=("name",),
            name="logical_vlan_group_name_no_scope",
            model_class=get_object_type_model("ipam.vlangroup"),
            condition=Q(scope_type__isnull=True),
        ),
    ],
    "wireless.wirelesslan": lambda: [
        ObjectMatchCriteria(
            fields=("ssid",),
            name="logical_wireless_lan_ssid_no_group_or_vlan",
            model_class=get_object_type_model("wireless.wirelesslan"),
            condition=Q(group__isnull=True, vlan__isnull=True),
        ),
        ObjectMatchCriteria(
            fields=("ssid", "group"),
            name="logical_wireless_lan_ssid_in_group",
            model_class=get_object_type_model("wireless.wirelesslan"),
            condition=Q(group__isnull=False),
        ),
        ObjectMatchCriteria(
            fields=("ssid", "vlan"),
            name="logical_wireless_lan_ssid_in_vlan",
            model_class=get_object_type_model("wireless.wirelesslan"),
            condition=Q(vlan__isnull=False),
        ),
    ],
    "virtualization.virtualmachine": lambda: [
        ObjectMatchCriteria(
            fields=("name",),
            name="logical_virtual_machine_name_no_cluster",
            model_class=get_object_type_model("virtualization.virtualmachine"),
            condition=Q(cluster__isnull=True),
        ),
    ],
    "ipam.service": lambda: [
        ObjectMatchCriteria(
            fields=("name",),
            name="logical_service_name_no_device_or_vm",
            model_class=get_object_type_model("ipam.service"),
            condition=Q(device__isnull=True, virtual_machine__isnull=True),
            max_version="4.2.99",
        ),
        ObjectMatchCriteria(
            fields=("name", "device"),
            name="logical_service_name_on_device",
            model_class=get_object_type_model("ipam.service"),
            condition=Q(device__isnull=False),
            max_version="4.2.99",
        ),
        ObjectMatchCriteria(
            fields=("name", "virtual_machine"),
            name="logical_service_name_on_vm",
            model_class=get_object_type_model("ipam.service"),
            condition=Q(virtual_machine__isnull=False),
            max_version="4.2.99",
        ),
        ObjectMatchCriteria(
            fields=("name", "parent_object_type", "parent_object_id"),
            name="logical_service_name_on_parent",
            model_class=get_object_type_model("ipam.service"),
            condition=Q(parent_object_type__isnull=False),
            min_version="4.3.0"
        ),
    ],
    "dcim.modulebay": lambda: [
        ObjectMatchCriteria(
            fields=("name", "device"),
            name="logical_module_bay_name_on_device",
            model_class=get_object_type_model("dcim.modulebay"),
        )
    ],
    "dcim.inventoryitem": lambda: [
        # TODO: this may be handleable by the existing constraints.
        # we ignore it due to null values for parent but could have
        # better coverage of this case perhaps.
        ObjectMatchCriteria(
            fields=("name", "device"),
            name="logical_inventory_item_name_on_device_no_parent",
            model_class=get_object_type_model("dcim.inventoryitem"),
            condition=Q(parent__isnull=True),
        )
    ],
    "ipam.fhrpgroup": lambda: [
        ObjectMatchCriteria(
            fields=("group_id",),
            name="logical_fhrp_group_id",
            model_class=get_object_type_model("ipam.fhrpgroup"),
        )
    ],
    "tenancy.contact": lambda: [
        ObjectMatchCriteria(
            # contacts are unconstrained in 4.3.0
            # in 4.2 they are constrained by unique name per group
            fields=("name", ),
            name="logical_contact_name",
            model_class=get_object_type_model("tenancy.contact"),
            min_version="4.3.0",
        )
    ],
    "dcim.devicerole": lambda: [
        ObjectMatchCriteria(
            fields=("name",),
            name="logical_device_role_name_no_parent",
            model_class=get_object_type_model("dcim.devicerole"),
            condition=Q(parent__isnull=True),
            min_version="4.3.0",
        ),
        ObjectMatchCriteria(
            fields=("slug",),
            name="logical_device_role_slug_no_parent",
            model_class=get_object_type_model("dcim.devicerole"),
            condition=Q(parent__isnull=True),
            min_version="4.3.0",
        )
    ],
    "extras.journalentry": lambda: [
        ObjectMatchCriteria(
            fields=("assigned_object_id", "assigned_object_type", "comments"),
            name="logical_journal_entry_assigned_object_comments",
            model_class=get_object_type_model("extras.journalentry"),
        )
    ],
}

@dataclass
class ObjectMatchCriteria:
    """
    Defines criteria for identifying a specific object.

    This matcher expects a fully 'transformed' and resolved
    set of fields. ie field names are snake case and match
    the model fields and any references to another object
    specify a specific id in the appropriate field name.
    eg device_id=123 etc and for any generic references,
    both the type and id should be specified, eg:
    scope_type="dcim.site" and scope_id=123
    """

    fields: tuple[str] | None = None
    expressions: tuple | None = None
    condition: Q | None = None
    model_class: type[models.Model] | None = None
    name: str | None = None

    min_version: str | None = None
    max_version: str | None = None

    def __hash__(self):
        """Hash the object match criteria."""
        return hash((self.fields, self.expressions, self.condition, self.model_class.__name__, self.name))

    def has_required_fields(self, data) -> bool:
        """Returns True if the data given contains a value for all fields referenced by the constraint."""
        return all(field in data for field in self._get_refs())

    @cache
    def _get_refs(self) -> set[str]:
        """Returns a set of all field names referenced by the constraint."""
        refs = set()
        if self.fields:
            refs.update(self.fields)
        elif self.expressions:
            for expr in self.expressions:
                refs |= _get_refs(expr)
        return frozenset(refs)

    @cache
    def _get_insensitive_refs(self) -> set[str]:
        """
        Returns a set of all field names that should be compared in a case insensitive manner.

        best effort, doesn't handle things being nested in a complex way.
        """
        refs = set()
        if self.expressions:
            for expr in self.expressions:
                # TODO be more careful here
                if expr.__class__.__name__ == "Lower":
                    for source_expr in getattr(expr, "source_expressions", []):
                        if hasattr(source_expr, "name"):
                            refs.add(source_expr.name)
        return refs

    def fingerprint(self, data: dict) -> str|None:
        """
        Returns a fingerprint of the data based on these criteria.

        These criteria that can be used to determine if two
        data structs roughly match.

        This is a best effort based on the referenced fields
        and some interrogation of case sensitivity. The
        real criteria are potentially complex...
        """
        if not self.has_required_fields(data):
            return None

        if self.condition:
            if not self._check_condition(data):
                return None

        # sort the fields by name
        sorted_fields = sorted(self._get_refs())
        insensitive = self._get_insensitive_refs()
        values = []
        for field in sorted_fields:
            value = data[field]
            if isinstance(value, dict):
                logger.warning(f"unexpected value type for fingerprinting: {value}")
                return None
            if field in insensitive:
                value = value.lower()
            values.append(value)

        return hash((self.model_class.__name__, self.name, tuple(values)))

    def _check_condition(self, data) -> bool:
        return self._check_condition_1(data, self.condition)

    def _check_condition_1(self, data, condition) -> bool:
        if condition is None:
            return True
        if isinstance(condition, tuple):
            return self._check_simple_condition(data, condition)

        if hasattr(condition, "connector") and condition.connector == Q.AND:
            result = True
            for child in condition.children:
                if not self._check_condition_1(data, child):
                    result = False
                    break
            if condition.negated:
                return not result
            return result
        # TODO handle OR ?
        logger.warning(f"Unhandled condition {condition}")
        return False

    def _check_simple_condition(self, data, condition) -> bool:
        if condition is None:
            return True

        k, v = condition
        result = False
        if k.endswith("__isnull"):
            k = k[:-8]
            is_null = k not in data or data[k] is None
            result = is_null == v
        else:
            result = k in data and data[k] == v

        return result

    def build_queryset(self, data) -> models.QuerySet:
        """Builds a queryset for the constraint with the given data."""
        if self.fields and len(self.fields) > 0:
            return self._build_fields_queryset(data)
        if self.expressions and len(self.expressions) > 0:
            return self._build_expressions_queryset(data)
        raise ValueError("No fields or expressions to build queryset from")

    def _build_fields_queryset(self, data) -> models.QuerySet: # noqa: C901
        """Builds a queryset for a simple set-of-fields constraint."""
        if not self._check_condition(data):
            return None

        data = self._prepare_data(data)
        lookup_kwargs = {}
        for field_name in self.fields:
            field = self.model_class._meta.get_field(field_name)
            if field_name not in data:
                return None  # cannot match, missing field data
            lookup_value = data.get(field_name)
            if isinstance(lookup_value, UnresolvedReference):
                return None  # cannot match, missing field data
            if isinstance(lookup_value, dict):
                return None  # cannot match, missing field data
            lookup_kwargs[field.name] = lookup_value

        qs = self.model_class.objects.filter(**lookup_kwargs)
        if self.condition:
            qs = qs.filter(self.condition)
        return qs

    def _build_expressions_queryset(self, data) -> models.QuerySet:
        """Builds a queryset for the constraint with the given data."""
        data = self._prepare_data(data)
        replacements = {
            F(field): Value(value) if isinstance(value, str | int | float | bool) else value
            for field, value in data.items()
        }

        filters = []
        for expr in self.expressions:
            if hasattr(expr, "get_expression_for_validation"):
                expr = expr.get_expression_for_validation()

            refs = [F(ref) for ref in _get_refs(expr)]
            for ref in refs:
                if ref not in replacements:
                    return None  # cannot match, missing field data
                if isinstance(replacements[ref], UnresolvedReference):
                    return None  # cannot match, missing field data

            rhs = expr.replace_expressions(replacements)
            condition = Exact(expr, rhs)
            filters.append(condition)

        qs = self.model_class.objects.filter(*filters)
        if self.condition:
            qs = qs.filter(self.condition)
        return qs

    def _prepare_data(self, data: dict) -> dict:
        prepared = {}
        for field_name, value in data.items():
            try:
                field = self.model_class._meta.get_field(field_name)
                # special handling for object type -> content type id
                if field.is_relation and hasattr(field, "related_model") and field.related_model == ContentType:
                    # Handle ManyToMany fields (list of object types) and ForeignKey fields (single object type)
                    if isinstance(value, list):
                        prepared[field_name] = [content_type_id(v) for v in value]
                    else:
                        prepared[field_name] = content_type_id(value)
                else:
                    prepared[field_name] = value

            except FieldDoesNotExist:
                continue
        return prepared



@dataclass
class CustomFieldMatcher:
    """A matcher for a unique custom field."""

    name: str
    custom_field: str
    model_class: type[models.Model]

    min_version: str | None = None
    max_version: str | None = None

    def fingerprint(self, data: dict) -> str|None:
        """Fingerprint the custom field value."""
        if not self.has_required_fields(data):
            return None

        value = data.get("custom_fields", {}).get(self.custom_field)
        if value is None:
            return None

        return hash((self.model_class.__name__, self.name, value))

    def build_queryset(self, data: dict) -> models.QuerySet:
        """Build a queryset for the custom field."""
        if not self.has_required_fields(data):
            return None

        value = data.get("custom_fields", {}).get(self.custom_field)
        if value is None:
            return None

        return self.model_class.objects.filter(**{f'custom_field_data__{self.custom_field}': value})

    def has_required_fields(self, data: dict) -> bool:
        """Returns True if the data given contains a value for all fields referenced by the constraint."""
        return self.custom_field in data.get("custom_fields", {})


@dataclass
class GlobalIPNetworkIPMatcher:
    """A matcher that ignores the mask."""

    ip_fields: tuple[str]
    vrf_field: str
    model_class: type[models.Model]
    name: str

    min_version: str | None = None
    max_version: str | None = None

    def _check_condition(self, data: dict) -> bool:
        """Check the condition for the custom field."""
        return data.get(self.vrf_field, None) is None

    def fingerprint(self, data: dict) -> str|None:
        """Fingerprint the custom field value."""
        if not self.has_required_fields(data):
            return None

        if not self._check_condition(data):
            return None

        values = []
        for field in self.ip_fields:
            value = self.ip_value(data, field)
            if value is None:
                return None
            values.append(value)

        return hash((self.model_class.__name__, self.name, tuple(values)))

    def has_required_fields(self, data: dict) -> bool:
        """Returns True if the data given contains a value for all fields referenced by the constraint."""
        return all(field in data for field in self.ip_fields)

    def ip_value(self, data: dict, field: str) -> str|None:
        """Get the IP value from the data."""
        value = data.get(field)
        if value is None:
            return None
        return _ip_only(value)

    def build_queryset(self, data: dict) -> models.QuerySet:
        """Build a queryset for the custom field."""
        if not self.has_required_fields(data):
            return None

        if not self._check_condition(data):
            return None

        filter = {
            f'{self.vrf_field}__isnull': True,
        }
        for field in self.ip_fields:
            value = self.ip_value(data, field)
            if value is None:
                return None
            filter[f'{field}__net_host'] = value

        return self.model_class.objects.filter(**filter)

@dataclass
class VRFIPNetworkIPMatcher:
    """Matches ip in a vrf, ignores mask."""

    ip_fields: tuple[str]
    vrf_field: str
    model_class: type[models.Model]
    name: str

    min_version: str | None = None
    max_version: str | None = None

    def _check_condition(self, data: dict) -> bool:
        """Check the condition for the custom field."""
        return data.get(self.vrf_field, None) is not None

    def fingerprint(self, data: dict) -> str|None:
        """Fingerprint the custom field value."""
        if not self.has_required_fields(data):
            return None

        if not self._check_condition(data):
            return None

        values = []
        for field in self.ip_fields:
            value = self.ip_value(data, field)
            if value is None:
                return None
            values.append(value)

        vrf_id = data[self.vrf_field]

        return hash((self.model_class.__name__, self.name, tuple(values), vrf_id))

    def has_required_fields(self, data: dict) -> bool:
        """Returns True if the data given contains a value for all fields referenced by the constraint."""
        return all(field in data for field in self.ip_fields) and self.vrf_field in data

    def ip_value(self, data: dict, field: str) -> str|None:
        """Get the IP value from the data."""
        value = data.get(field)
        if value is None:
            return None
        return _ip_only(value)

    def build_queryset(self, data: dict) -> models.QuerySet:
        """Build a queryset for the custom field."""
        if not self.has_required_fields(data):
            return None

        if not self._check_condition(data):
            return None

        filter = {}
        for field in self.ip_fields:
            value = self.ip_value(data, field)
            if value is None:
                return None
            filter[f'{field}__net_host'] = value

        vrf_id = data[self.vrf_field]
        if isinstance(vrf_id, UnresolvedReference):
            return None
        filter[f'{self.vrf_field}'] = vrf_id

        return self.model_class.objects.filter(**filter)


def _ip_only(value: str) -> str|None:
    try:
        ip = netaddr.IPNetwork(value)
        value = ip.ip
    except netaddr.core.AddrFormatError:
        return None

    return value

@dataclass
class AutoSlugMatcher:
    """A special matcher that tries to match on auto generated slugs."""

    name: str
    slug_field: str
    model_class: type[models.Model]

    min_version: str | None = None
    max_version: str | None = None

    def fingerprint(self, data: dict) -> str|None:
        """Fingerprint the custom field value."""
        if not self.has_required_fields(data):
            return None

        slug = data.get('_auto_slug', None)
        if slug is None:
            return None

        return hash((self.model_class.__name__, self.name, slug.value))

    def build_queryset(self, data: dict) -> models.QuerySet:
        """Build a queryset for the custom field."""
        if not self.has_required_fields(data):
            return None

        slug = data.get('_auto_slug', None)
        if slug is None:
            return None

        return self.model_class.objects.filter(**{f'{self.slug_field}': str(slug.value)})

    def has_required_fields(self, data: dict) -> bool:
        """Returns True if the data given contains a value for all fields referenced by the constraint."""
        return '_auto_slug' in data


def get_model_matchers(model_class) -> list:
    """Extract unique constraints from a Django model."""
    matchers = []
    matchers += _get_model_matchers(model_class)

    # TODO(ltucker): this should also be cacheable, but we need a signal to invalidate
    if hasattr(model_class, "get_custom_fields"):
        unique_custom_fields = CustomField.objects.get_for_model(model_class).filter(unique=True)
        if unique_custom_fields:
            for cf in unique_custom_fields:
                matchers.append(
                    CustomFieldMatcher(
                        model_class=model_class,
                        custom_field=cf.name,
                        name=f"unique_custom_field_{cf.name}",
                    )
                )
    matchers += _get_autoslug_matchers(model_class)
    return matchers

@lru_cache(maxsize=256)
def _get_autoslug_matchers(model_class) -> list:
    matchers = []
    for field in model_class._meta.fields:
        if isinstance(field, SlugField):
            matchers.append(
                AutoSlugMatcher(
                    model_class=model_class,
                    slug_field=field.name,
                    name=f"unique_autoslug_{field.name}",
                )
            )
            break
    return matchers

@lru_cache(maxsize=256)
def _get_model_matchers(model_class) -> list[ObjectMatchCriteria]:
    object_type = get_object_type(model_class)
    matchers = [
        x for x in _LOGICAL_MATCHERS.get(object_type, lambda: [])()
        if in_version_range(x.min_version, x.max_version)
    ]

    # collect single fields that are unique
    for field in model_class._meta.fields:
        if field.name == "id":
            # TODO(ltucker): more django-general detection of pk field?
            continue

        if field.unique:
            matchers.append(
                ObjectMatchCriteria(
                    model_class=model_class,
                    fields=(field.name,),
                    name=f"unique_{field.name}",
                )
            )

    # collect UniqueConstraint constraints
    for constraint in model_class._meta.constraints:
        if not _is_supported_constraint(constraint, model_class):
            continue
        if len(constraint.fields) > 0:
            matchers.append(
                ObjectMatchCriteria(
                    model_class=model_class,
                    fields=tuple(constraint.fields),
                    condition=constraint.condition,
                    name=constraint.name,
                )
            )
        elif len(constraint.expressions) > 0:
            matchers.append(
                ObjectMatchCriteria(
                    model_class=model_class,
                    expressions=tuple(constraint.expressions),
                    condition=constraint.condition,
                    name=constraint.name,
                )
            )
        else:
            logger.debug(
                f"Constraint {constraint.name} on {model_class.__name__} had no fields or expressions (skipped)"
            )
            # (this shouldn't happen / enforced by django)
            continue

    return matchers


def _is_supported_constraint(constraint, model_class) -> bool:
    if not isinstance(constraint, models.UniqueConstraint):
        return False

    if len(constraint.opclasses) > 0:
        logger.warning(f"Constraint {constraint.name} on {model_class.__name__} had opclasses (skipped)")
        return False

    if constraint.nulls_distinct is not None and constraint.nulls_distinct is True:
        logger.warning(f"Constraint {constraint.name} on {model_class.__name__} had nulls_distinct (skipped)")
        return False

    for field_name in constraint.fields:
        field = model_class._meta.get_field(field_name)
        if field.generated:
            logger.warning(
                f"Constraint {constraint.name} on {model_class.__name__} had"
                f" generated field {field_name} (skipped)"
            )
            return False

    return True

def _get_refs(expr) -> set[str]:
    refs = set()
    if isinstance(expr, str):
        refs.add(expr)
    elif isinstance(expr, F):
        refs.add(expr.name)
    elif hasattr(expr, "get_source_expressions"):
        for subexpr in expr.get_source_expressions():
            refs |= _get_refs(subexpr)
    else:
        logger.warning(f"Unhandled expression type for _get_refs: {type(expr)}")
    return refs

def _fingerprint_all(data: dict, object_type: str|None = None) -> str:
    """
    Returns a fingerprint of the data based on all fields.

    Data should be a (flattened) dictionary of field values.
    This ignores any fields that start with an underscore.
    """
    if data is None:
        return None

    try:
        values = ["object_type", object_type]
        for k, v in sorted(data.items()):
            if k.startswith("_"):
                continue
            values.append(k)
            if isinstance(v, list | tuple):
                values.extend(sorted(_as_tuples(v)))
            elif isinstance(v, dict):
                values.append(_fingerprint_all(v))
            else:
                values.append(v)

        return hash(tuple(values))
    except Exception as e:
        logger.error(f"Error fingerprinting data: {e}")
        raise

def _as_tuples(vs):
    if isinstance(vs, list):
        return tuple(_as_tuples(v) for v in vs)
    if isinstance(vs, dict):
        return tuple((k, _as_tuples(v)) for k, v in vs.items())
    return vs

def fingerprints(data: dict, object_type: str) -> list[str]:
    """
    Get fingerprints for a data structure.

    This returns all fingerprints for the given data that
    have required fields.
    """
    if data is None:
        return None

    model_class = get_object_type_model(object_type)
    # check any known match criteria
    fps = []
    for matcher in get_model_matchers(model_class):
        fp = matcher.fingerprint(data)
        if fp is not None:
            fps.append(fp)
    fp = _fingerprint_all(data, object_type)
    fps.append(fp)
    return fps

def find_existing_object(data: dict, object_type: str): # noqa: C901
    """
    Find an existing object that matches the given data.

    Uses all object match criteria to look for an existing
    object. Returns the first match found.

    Returns the object if found, otherwise None.
    """
    model_class = get_object_type_model(object_type)
    for matcher in get_model_matchers(model_class):
        if not matcher.has_required_fields(data):
            continue
        q = matcher.build_queryset(data)
        if q is None:
            continue
        existing = q.order_by('pk').first()
        if existing is not None:
            return existing
    return None



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - Database migrations."""

from django.db import migrations, models


class Migration(migrations.Migration):
    """Add optional branch_id field to Setting model."""

    dependencies = [
        ("netbox_diode_plugin", "0007_setting_model_cleanup"),
    ]

    operations = [
        migrations.AddField(
            model_name="setting",
            name="branch_id",
            field=models.BigIntegerField(
                blank=True,
                null=True,
                help_text="ID of the branch for NetBox Branching plugin integration",
            ),
        ),
    ]



#!/usr/bin/env python
# Copyright 2024 NetBox Labs Inc
"""Diode NetBox Plugin - Tables."""
import logging

import django_tables2 as tables
from django.urls import reverse
from django.utils.dateparse import parse_datetime
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from netbox.tables import BaseTable, columns


class ClientCredentialsTable(BaseTable):
    """Client credentials table."""

    label = tables.Column(
        verbose_name=_("Name"),
        accessor="client_name",
        orderable=False,
    )
    client_id = tables.Column(
        verbose_name=_("Client ID"),
        accessor="client_id",
        orderable=False,
    )
    created_at = columns.DateTimeColumn(
        verbose_name=_("Created"),
        accessor="created_at",
        orderable=False,
    )
    client_secret = tables.Column(
        verbose_name=_("Client Secret"),
        empty_values=(),
        orderable=False,
    )
    actions = tables.Column(
        verbose_name=_(""),
        orderable=False,
        empty_values=(),
        attrs={
            "td": {
                "class": "text-end",
            }
        },
    )

    exempt_columns = ("actions")
    embedded = False

    class Meta:
        """Meta class."""

        attrs = {
            "class": "table table-hover object-list",
            "td": {"class": "align-middle"},
        }
        fields = None
        default_columns = (
            "label",
            "client_id",
            "created_at",
            "client_secret",
            "actions",
        )

        empty_text = _("No Client Credentials to display")
        footer = False

    def render_client_secret(self, value):
        """Render client secret."""
        return "*****"

    def render_created_at(self, value):
        """Render created at."""
        if value:
            return parse_datetime(value)
        return "-"

    def render_actions(self, record):
        """Render actions."""
        delete_url = reverse(
            "plugins:netbox_diode_plugin:client_credential_delete",
            kwargs={"client_credential_id": record["client_id"]},
        )

        buttons = f"""
            <a class="btn btn-sm btn-red"
                data-bs-toggle="tooltip"
                data-bs-placement="top"
                title="Delete"
                href="{delete_url}"
                type="button"
                aria-label="Delete">
                <i class="mdi mdi-trash-can-outline"></i>
            </a>
        """  # noqa: E501

        return mark_safe(buttons)



#!/usr/bin/env python
# Copyright 2025 NetBox Labs, Inc.
"""Diode NetBox Plugin - API URLs."""

from django.urls import include, path
from netbox.api.routers import NetBoxRouter

from .views import ApplyChangeSetView, GenerateDiffView, GetDefaultBranchView

router = NetBoxRouter()

urlpatterns = [
    path("apply-change-set/", ApplyChangeSetView.as_view()),
    path("generate-diff/", GenerateDiffView.as_view()),
    path("default-branch/", GetDefaultBranchView.as_view()),
    path("", include(router.urls)),
]
